package java.util;

import java.security.PrivilegedAction;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.BufferedReader;
import java.net.URL;
import java.security.AccessController;
import java.security.AccessControlContext;

public final class ServiceLoader<S> implements Iterable<S>
{
    private static final String PREFIX = "META-INF/services/";
    private final Class<S> service;
    private final ClassLoader loader;
    private final AccessControlContext acc;
    private LinkedHashMap<String, S> providers;
    private ServiceLoader.LazyIterator lookupIterator;
    
    public void reload() {
        this.providers.clear();
        this.lookupIterator = new ServiceLoader.LazyIterator(this.service, this.loader);
    }
    
    private ServiceLoader(Class<S> clazz, ClassLoader classLoader) {
        this.providers = new LinkedHashMap();
        this.service = (Class<S>)Objects.requireNonNull(clazz, "Service interface cannot be null");
        this.loader = ((classLoader == null) ? ClassLoader.getSystemClassLoader() : classLoader);
        this.acc = ((System.getSecurityManager() != null) ? AccessController.getContext() : null);
        this.reload();
    }
    
    private static void fail(Class<?> clazz, String str, Throwable t) throws ServiceConfigurationError {
        throw new ServiceConfigurationError(clazz.getName() + ": " + str, t);
    }
    
    private static void fail(Class<?> clazz, String str) throws ServiceConfigurationError {
        throw new ServiceConfigurationError(clazz.getName() + ": " + str);
    }
    
    private static void fail(Class<?> clazz, URL obj, int i, String str) throws ServiceConfigurationError {
        fail(clazz, obj + ":" + i + ": " + str);
    }
    
    private int parseLine(Class<?> clazz, URL url, BufferedReader bufferedReader, int n, List<String> list) throws IOException, ServiceConfigurationError {
        String s = bufferedReader.readLine();
        if (s == null) {
            return -1;
        }
        int index = s.indexOf(35);
        if (index >= 0) {
            s = s.substring(0, index);
        }
        String trim = s.trim();
        int length = trim.length();
        if (length != 0) {
            if (trim.indexOf(32) >= 0 || trim.indexOf(9) >= 0) {
                fail(clazz, url, n, "Illegal configuration-file syntax");
            }
            int codePoint = trim.codePointAt(0);
            if (!Character.isJavaIdentifierStart(codePoint)) {
                fail(clazz, url, n, "Illegal provider-class name: " + trim);
            }
            int codePoint2;
            for (int i = Character.charCount(codePoint); i < length; i += Character.charCount(codePoint2)) {
                codePoint2 = trim.codePointAt(i);
                if (!Character.isJavaIdentifierPart(codePoint2) && codePoint2 != 46) {
                    fail(clazz, url, n, "Illegal provider-class name: " + trim);
                }
            }
            if (!this.providers.containsKey(trim) && !list.contains(trim)) {
                list.add(trim);
            }
        }
        return n + 1;
    }
    
    private Iterator<String> parse(Class<?> clazz, URL url) throws ServiceConfigurationError {
        InputStream openStream = null;
        BufferedReader bufferedReader = null;
        ArrayList<String> list = new ArrayList();
        try {
            openStream = url.openStream();
            bufferedReader = new BufferedReader(new InputStreamReader(openStream, "utf-8"));
            int line = 1;
            while ((line = this.parseLine(clazz, url, bufferedReader, line, list)) >= 0) {}
            try {
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (openStream != null) {
                    openStream.close();
                }
            }
            catch (IOException ex) {
                fail(clazz, "Error closing configuration file", ex);
            }
        }
        catch (IOException ex2) {
            fail(clazz, "Error reading configuration file", ex2);
            try {
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (openStream != null) {
                    openStream.close();
                }
            }
            catch (IOException ex3) {
                fail(clazz, "Error closing configuration file", ex3);
            }
        }
        finally {
            try {
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (openStream != null) {
                    openStream.close();
                }
            }
            catch (IOException ex4) {
                fail(clazz, "Error closing configuration file", ex4);
            }
        }
        return list.iterator();
    }
    
    @Override
    public Iterator<S> iterator() {
        return new Iterator<S>() {
            Iterator<Map.Entry<String, S>> knownProviders = ServiceLoader.this.providers.entrySet().iterator();
            
            @Override
            public boolean hasNext() {
                return this.knownProviders.hasNext() || ServiceLoader.this.lookupIterator.hasNext();
            }
            
            @Override
            public S next() {
                if (this.knownProviders.hasNext()) {
                    return ((Map.Entry)this.knownProviders.next()).getValue();
                }
                return ServiceLoader.this.lookupIterator.next();
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    public static <S> ServiceLoader<S> load(Class<S> clazz, ClassLoader classLoader) {
        return new ServiceLoader(clazz, classLoader);
    }
    
    public static <S> ServiceLoader<S> load(Class<S> clazz) {
        return load(clazz, Thread.currentThread().getContextClassLoader());
    }
    
    public static <S> ServiceLoader<S> loadInstalled(Class<S> clazz) {
        ClassLoader classLoader = ClassLoader.getSystemClassLoader();
        ClassLoader classLoader2 = null;
        while (classLoader != null) {
            classLoader2 = classLoader;
            classLoader = classLoader.getParent();
        }
        return load(clazz, classLoader2);
    }
    
    @Override
    public String toString() {
        return "java.util.ServiceLoader[" + this.service.getName() + "]";
    }
    
    static void access$100(Class clazz, String s, Throwable t) throws ServiceConfigurationError {
        fail(clazz, s, t);
    }
    
    static Iterator access$200(ServiceLoader serviceLoader, Class clazz, URL url) throws ServiceConfigurationError {
        return serviceLoader.parse(clazz, url);
    }
    
    static void access$300(Class clazz, String s) throws ServiceConfigurationError {
        fail(clazz, s);
    }
    
    private class LazyIterator implements Iterator<S>
    {
        Class<S> service;
        ClassLoader loader;
        Enumeration<URL> configs;
        Iterator<String> pending;
        String nextName;
        
        private LazyIterator(Class<S> service, ClassLoader loader) {
            this.configs = null;
            this.pending = null;
            this.nextName = null;
            this.service = service;
            this.loader = loader;
        }
        
        private boolean hasNextService() {
            if (this.nextName != null) {
                return true;
            }
            if (this.configs == null) {
                try {
                    String string = "META-INF/services/" + this.service.getName();
                    if (this.loader == null) {
                        this.configs = ClassLoader.getSystemResources(string);
                    }
                    else {
                        this.configs = this.loader.getResources(string);
                    }
                }
                catch (IOException ex) {
                    ServiceLoader.access$100(this.service, "Error locating configuration files", ex);
                }
            }
            while (this.pending == null || !this.pending.hasNext()) {
                if (!this.configs.hasMoreElements()) {
                    return false;
                }
                this.pending = ServiceLoader.access$200(ServiceLoader.this, this.service, (URL)this.configs.nextElement());
            }
            this.nextName = (String)this.pending.next();
            return true;
        }
        
        private S nextService() {
            if (!this.hasNextService()) {
                throw new NoSuchElementException();
            }
            String nextName = this.nextName;
            this.nextName = null;
            Class<Object> forName = null;
            try {
                forName = Class.forName(nextName, false, this.loader);
            }
            catch (ClassNotFoundException ex) {
                ServiceLoader.access$300(this.service, "Provider " + nextName + " not found");
            }
            if (!this.service.isAssignableFrom(forName)) {
                ServiceLoader.access$300(this.service, "Provider " + nextName + " not a subtype");
            }
            try {
                Object cast = this.service.cast(forName.newInstance());
                ServiceLoader.this.providers.put(nextName, cast);
                return cast;
            }
            catch (Throwable t) {
                ServiceLoader.access$100(this.service, "Provider " + nextName + " could not be instantiated", t);
                throw new Error();
            }
        }
        
        @Override
        public boolean hasNext() {
            if (ServiceLoader.this.acc == null) {
                return this.hasNextService();
            }
            return (boolean)AccessController.doPrivileged(new PrivilegedAction<Boolean>() {
                @Override
                public Boolean run() {
                    return Boolean.valueOf(ServiceLoader.LazyIterator.access$600(ServiceLoader.LazyIterator.this));
                }
            }, ServiceLoader.this.acc);
        }
        
        @Override
        public S next() {
            if (ServiceLoader.this.acc == null) {
                return this.nextService();
            }
            return AccessController.doPrivileged(new PrivilegedAction<S>() {
                @Override
                public S run() {
                    return ServiceLoader.LazyIterator.access$700(ServiceLoader.LazyIterator.this);
                }
            }, ServiceLoader.this.acc);
        }
        
        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
        
        static boolean access$600(ServiceLoader.LazyIterator lazyIterator) {
            return lazyIterator.hasNextService();
        }
        
        static Object access$700(ServiceLoader.LazyIterator lazyIterator) {
            return lazyIterator.nextService();
        }
    }
}
